/** @file ledbox.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.08.08 -

	@brief [ led box ]
*/
#include "ledbox.h"
#include "../emu.h"
#include <windowsx.h>

#define RECT_IN(st, xx, yy, ww, hh) { \
	st.pt.x = xx; \
	st.pt.y = yy; \
	st.sz.cx = ww; \
	st.sz.cy = hh; \
}

#define POINT_IN(st, xx, yy, sf) { \
	st.pt.x = xx; \
	st.pt.y = yy; \
}

LedBoxBase::LedBoxBase() : DibSection()
{
	hInstance = NULL;

	for(int i=0; i<LED_TYPE_END; i++) {
		ZeroMemory(&led[i], sizeof(led_t));
		led[i].hBmp = NULL;
		led[i].hDC = NULL;
	}
	ZeroMemory(led_pt, sizeof(led_pt));

	flag_old = -1;
	dist.x = 0; dist.y = 0;
	mode = 0;
	prev_mode = -1;
	dist_set[0].x = 2; dist_set[0].y = 2;
	dist_set[1].x = 2; dist_set[1].y = 2;
	memset(&win_pt, 0, sizeof(win_pt));
	memset(&parent_pt, 0, sizeof(parent_pt));

	visible = false;
	inside = false;
}

LedBoxBase::~LedBoxBase()
{
	for(int i=0; i<LED_TYPE_END; i++) {
		if (led[i].hDC) DeleteDC(led[i].hDC);
		if (led[i].hBmp) DeleteObject(led[i].hBmp);
	}
}

// ʏ
bool LedBoxBase::InitScreen(HDC hdc)
{
	hInstance = ::GetModuleHandle(NULL);

	enable = false;

	// LEDwi
	if (!create_bitmap(hdc, &led[LED_TYPE_BASE], IDB_KBLEDBOX)) return enable;

	parent_pt.w = led[LED_TYPE_BASE].sz.cx;
	parent_pt.h = led[LED_TYPE_BASE].sz.cy;

	// Dib쐬
	if (!CreateDib(hdc, parent_pt.w, parent_pt.h)) return enable;

	// LEDp[c쐬
	if (!create_bitmap(hdc, &led[LED_TYPE_PARTS], IDB_LED_PARTS)) return enable;

	// horiz
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 7 * 16, 0, 13, 6, &led[LED_TYPE_HR])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 9 * 16, 0, 13, 6, &led[LED_TYPE_HG])) return enable;
	// vert
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 1 * 16, 0, 6, 13, &led[LED_TYPE_VR])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 3 * 16, 0, 6, 13, &led[LED_TYPE_VG])) return enable;
	// datarec
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 11 * 16, 0, 13, 11, &led[LED_TYPE_RR])) return enable;	// play led
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 13 * 16, 0, 11, 11, &led[LED_TYPE_CL])) return enable;	// rec led
	// digit
	if (!create_bitmap(hdc, &led[LED_TYPE_DIGIT], IDB_DIGIT_10)) return enable;
	// tape icon
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 14 * 16, 0, 15, 11, &led[LCD_TYPE_TAPE])) return enable;
	// fdd icon
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 15 * 16, 0, 13, 13, &led[LCD_TYPE_FDD0])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 16 * 16, 0, 13, 13, &led[LCD_TYPE_FDD1])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 17 * 16, 0, 13, 13, &led[LCD_TYPE_FDD2])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 18 * 16, 0, 13, 13, &led[LCD_TYPE_FDD3])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 21 * 16, 0, 7, 17, &led[LCD_TYPE_3_FDD35H])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 22 * 16, 0, 7, 17, &led[LCD_TYPE_5_FDD35H])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 23 * 16, 0, 7, 17, &led[LCD_TYPE_H_FDD35H])) return enable;
	// ab mode
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 24 * 16, 0, 10, 18, &led[LCD_TYPE_AB_A])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 25 * 16, 0, 10, 18, &led[LCD_TYPE_AB_B])) return enable;
	// 1,2mhz
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 26 * 16, 0, 10, 17, &led[LCD_TYPE_12M_2])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 27 * 16, 0, 10, 17, &led[LCD_TYPE_12M_1])) return enable;
	// s,u
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 28 * 16, 0, 10, 17, &led[LCD_TYPE_SU_S])) return enable;
	if (!create_parts(hdc, led[LED_TYPE_PARTS], 29 * 16, 0, 10, 17, &led[LCD_TYPE_SU_U])) return enable;

	// ʒuݒ
	POINT_IN(led_pt[LED_POS_BASE], 0, 0, led[LED_TYPE_BASE]);
	// keyboard led
	POINT_IN(led_pt[LED_POS_KATAKANA], 31, 3, led[LED_TYPE_HR]);
	POINT_IN(led_pt[LED_POS_HIRAGANA], 31, 11, led[LED_TYPE_HG]);
	POINT_IN(led_pt[LED_POS_CAPS], 3, 3, led[LED_TYPE_VR]);
	// datarec led
	POINT_IN(led_pt[LED_POS_PLAY], 65, 4, led[LED_TYPE_RR]);
	POINT_IN(led_pt[LCD_POS_PLAY], 78, 4, led[LCD_TYPE_TAPE]);
	POINT_IN(led_pt[LED_POS_REC],  96, 4, led[LED_TYPE_CL]);
	POINT_IN(led_pt[LCD_POS_REC], 108, 4, led[LCD_TYPE_TAPE]);
	// fdd type
	POINT_IN(led_pt[LCD_POS_FDD35H], 163, 1, led[LCD_TYPE_3_FDD35H]);
	// fdd 0
	POINT_IN(led_pt[LED_POS_FDD0], 173, 3, led[LED_TYPE_VR]);
	POINT_IN(led_pt[LCD_POS_FDD0], 180, 3, led[LCD_TYPE_FDD0]);
	// fdd 1
	POINT_IN(led_pt[LED_POS_FDD1], 198, 3, led[LED_TYPE_VR]);
	POINT_IN(led_pt[LCD_POS_FDD1], 205, 3, led[LCD_TYPE_FDD1]);
	// fdd 2
	POINT_IN(led_pt[LED_POS_FDD2], 223, 3, led[LED_TYPE_VR]);
	POINT_IN(led_pt[LCD_POS_FDD2], 230, 3, led[LCD_TYPE_FDD2]);
	// fdd 3
	POINT_IN(led_pt[LED_POS_FDD3], 248, 3, led[LED_TYPE_VR]);
	POINT_IN(led_pt[LCD_POS_FDD3], 255, 3, led[LCD_TYPE_FDD3]);
	// digit
	RECT_IN(led_pt[LED_POS_DIGIT1], 147, 4, 8, 12);	// 1
	RECT_IN(led_pt[LED_POS_DIGIT2], 138, 4, 8, 12);	// 10
	RECT_IN(led_pt[LED_POS_DIGIT3], 129, 4, 8, 12);	// 100
	// ab mode
	POINT_IN(led_pt[LCD_POS_AB], 276, 1, led[LCD_TYPE_AB_A]);
	// 1,2mhz
	POINT_IN(led_pt[LCD_POS_12M], 288, 2, led[LCD_TYPE_12M_1]);
	// s,u
	POINT_IN(led_pt[LCD_POS_SU], 298, 2, led[LCD_TYPE_SU_S]);

	//
	enable = true;
	copy_led(led_pt[LED_POS_BASE], led[LED_TYPE_BASE]);
	return enable;
}

bool LedBoxBase::create_bitmap(HDC hdc, led_t *l, DWORD id)
{
	BITMAPINFO bInfo;
	ZeroMemory(&bInfo, sizeof(bInfo));
	bInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

	l->hBmp = LoadBitmap(hInstance,MAKEINTRESOURCE(id));
	if (l->hBmp == NULL) return false;
	l->hDC = CreateCompatibleDC(hdc);
	if (l->hDC == NULL) return false;
	SelectObject(l->hDC, l->hBmp);
	GetDIBits(hdc, l->hBmp, 0, 1, NULL, &bInfo, DIB_RGB_COLORS);
	l->sz.cx = bInfo.bmiHeader.biWidth;
	l->sz.cy = bInfo.bmiHeader.biHeight;

	return true;
}

bool LedBoxBase::create_parts(HDC hdc, const led_t &srcl, int x, int y, int w, int h, led_t *l)
{
	BITMAPINFO bInfo;
	ZeroMemory(&bInfo, sizeof(bInfo));
	bInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	GetDIBits(hdc, srcl.hBmp, 0, 1, NULL, &bInfo, DIB_RGB_COLORS);
	bInfo.bmiHeader.biWidth = w;
	bInfo.bmiHeader.biHeight = h;

	l->hBmp = CreateDIBitmap(hdc, &bInfo.bmiHeader, 0, NULL, NULL, DIB_RGB_COLORS);
	if (l->hBmp == NULL) return false;
	l->hDC = CreateCompatibleDC(hdc);
	if (l->hDC == NULL) return false;
	SelectObject(l->hDC, l->hBmp);
	l->sz.cx = bInfo.bmiHeader.biWidth;
	l->sz.cy = bInfo.bmiHeader.biHeight;

	BitBlt(l->hDC, 0, 0, l->sz.cx, l->sz.cy, srcl.hDC, x, y, SRCCOPY);

	return true;
}

void LedBoxBase::Show(int flag)
{
	visible = ((flag & 1) != 0);
	inside = ((flag & 8) != 0);

	show_dialog();
}

void LedBoxBase::Update(uint64 flag)
{
	if (!enable) return;

	if (flag != flag_old) {
		copy_led(led_pt[LED_POS_BASE], led[LED_TYPE_BASE]);

		// katakana
		if (flag & 0x0001) {
			copy_led(led_pt[LED_POS_KATAKANA], led[LED_TYPE_HR]);
		}
		// hiragana
		if (flag & 0x0002) {
			copy_led(led_pt[LED_POS_HIRAGANA], led[LED_TYPE_HG]);
		}
		// caps
		if (!(flag & 0x0004)) {
			copy_led(led_pt[LED_POS_CAPS], led[LED_TYPE_VR]);
		}
		// play
		if (flag & 0x0010) {
			copy_led(led_pt[LED_POS_PLAY], led[LED_TYPE_RR]);
		}
		// rec
		if (flag & 0x0020) {
			copy_led(led_pt[LED_POS_REC], led[LED_TYPE_CL]);
		}
		// play lcd icon
		if (flag & 0x0040) {
			copy_led(led_pt[LCD_POS_PLAY], led[LCD_TYPE_TAPE]);
		}
		// rec lcd icon
		if (flag & 0x0080) {
			copy_led(led_pt[LCD_POS_REC], led[LCD_TYPE_TAPE]);
		}

		// fdd type
		if ((flag & 0x3000000000) == 0x1000000000) {
			copy_led(led_pt[LCD_POS_FDD35H], led[LCD_TYPE_3_FDD35H]);
		} else if ((flag & 0x3000000000) == 0x2000000000) {
			copy_led(led_pt[LCD_POS_FDD35H], led[LCD_TYPE_5_FDD35H]);
		} else if (flag & 0x3000000000) {
			copy_led(led_pt[LCD_POS_FDD35H], led[LCD_TYPE_H_FDD35H]);
		}

		// fdd0
		if ((flag & 0x11000000) == 0x01000000) {
			copy_led(led_pt[LED_POS_FDD0], led[LED_TYPE_VG]);
		} else if ((flag & 0x11000000) == 0x11000000) {
			copy_led(led_pt[LED_POS_FDD0], led[LED_TYPE_VR]);
		}
		if (flag & 0x0100000000) {
			copy_led(led_pt[LCD_POS_FDD0], led[LCD_TYPE_FDD0]);
		}
		// fdd1
		if ((flag & 0x22000000) == 0x02000000) {
			copy_led(led_pt[LED_POS_FDD1], led[LED_TYPE_VG]);
		} else if ((flag & 0x22000000) == 0x22000000) {
			copy_led(led_pt[LED_POS_FDD1], led[LED_TYPE_VR]);
		}
		if (flag & 0x0200000000) {
			copy_led(led_pt[LCD_POS_FDD1], led[LCD_TYPE_FDD1]);
		}
		// fdd2
		if ((flag & 0x44000000) == 0x04000000) {
			copy_led(led_pt[LED_POS_FDD2], led[LED_TYPE_VG]);
		} else if ((flag & 0x44000000) == 0x44000000) {
			copy_led(led_pt[LED_POS_FDD2], led[LED_TYPE_VR]);
		}
		if (flag & 0x0400000000) {
			copy_led(led_pt[LCD_POS_FDD2], led[LCD_TYPE_FDD2]);
		}
		// fdd3
		if ((flag & 0x88000000) == 0x08000000) {
			copy_led(led_pt[LED_POS_FDD3], led[LED_TYPE_VG]);
		} else if ((flag & 0x88000000) == 0x88000000) {
			copy_led(led_pt[LED_POS_FDD3], led[LED_TYPE_VR]);
		}
		if (flag & 0x0800000000) {
			copy_led(led_pt[LCD_POS_FDD3], led[LCD_TYPE_FDD3]);
		}

		// abmode
		if (flag & 0x10000000000) {
			copy_led(led_pt[LCD_POS_AB], led[LCD_TYPE_AB_A]);
		} else {
			copy_led(led_pt[LCD_POS_AB], led[LCD_TYPE_AB_B]);
		}
		// 1,2mhz
		if (flag & 0x20000000000) {
			copy_led(led_pt[LCD_POS_12M], led[LCD_TYPE_12M_2]);
		} else {
			copy_led(led_pt[LCD_POS_12M], led[LCD_TYPE_12M_1]);
		}
		// s,u
		if (flag & 0x40000000000) {
			copy_led(led_pt[LCD_POS_SU], led[LCD_TYPE_SU_S]);
		} else {
			copy_led(led_pt[LCD_POS_SU], led[LCD_TYPE_SU_U]);
		}

		// cmt counter
		uint32 count = ((flag >> 8) & 0xffff);
		int i = (count / 12);
		int f = (count % 12);
		POINT pt={0,0};
		pt.y = (i % 10) * led_pt[LED_POS_DIGIT1].sz.cy + f;
		copy_led_s(led_pt[LED_POS_DIGIT1], led[LED_TYPE_DIGIT], pt);
		pt.y = ((i / 10) % 10) * led_pt[LED_POS_DIGIT2].sz.cy + ((i % 10) == 9 ? f : 0);
		copy_led_s(led_pt[LED_POS_DIGIT2], led[LED_TYPE_DIGIT], pt);
		pt.y = ((i / 100) % 10) * led_pt[LED_POS_DIGIT3].sz.cy + ((i % 100) == 99 ? f : 0);
		copy_led_s(led_pt[LED_POS_DIGIT3], led[LED_TYPE_DIGIT], pt);

		if (visible && !inside) {
			need_update_dialog();
		}
	}

	flag_old = flag;
}

void LedBoxBase::copy_led(led_pt_t &p, led_t &l)
{
	BitBlt(hMainDib, p.pt.x, p.pt.y, l.sz.cx, l.sz.cy, l.hDC, 0, 0, SRCCOPY);
}

void LedBoxBase::copy_led_s(led_pt_t &p, led_t &l, POINT &pt)
{
	BitBlt(hMainDib, p.pt.x, p.pt.y, p.sz.cx, p.sz.cy, l.hDC, pt.x, pt.y, SRCCOPY);
}

void LedBoxBase::SetPos(int left, int top, int right, int bottom, int place)
{
	win_pt.left = left;
	win_pt.top = top;
	win_pt.right = right;
	win_pt.bottom = bottom;

	if (place & 1) {
		// base right
		parent_pt.x = right - parent_pt.w;
	} else {
		// base left
		parent_pt.x = left;
	}
	if (place & 2) {
		// base bottom
		parent_pt.y = bottom - parent_pt.h;
	} else {
		// base top
		parent_pt.y = top;
	}

	move_in_place(place);

	win_pt.place = place;
}

void LedBoxBase::SetPos(int place)
{
	SetPos(win_pt.left, win_pt.top, win_pt.right, win_pt.bottom, place);
}

void LedBoxBase::Draw(HDC hdc)
{
	if (!visible || !inside || !enable) return;

	// copy to main context
	BitBlt(hdc, parent_pt.x, parent_pt.y, parent_pt.w, parent_pt.h, hMainDib, 0, 0, SRCCOPY);
}

void LedBoxBase::Draw(LPDIRECT3DSURFACE9 suf)
{
	if (!visible || !inside || !enable || !suf) return;

	HDC hdc;
	HRESULT hre = suf->GetDC(&hdc);
	if (hre == D3D_OK) {
		// copy to main context
		BitBlt(hdc, parent_pt.x, parent_pt.y, parent_pt.w, parent_pt.h, hMainDib, 0, 0, SRCCOPY);
	}
	suf->ReleaseDC(hdc);
}

void LedBoxBase::SetMode(int val)
{
	val = (val < 0) ? 0 : (val > 1 ? 1 : val);
	dist_set[mode].x = dist.x;
	dist_set[mode].y = dist.y;
	prev_mode = mode;
	mode = val;
	dist.x = dist_set[mode].x;
	dist.y = dist_set[mode].y;
}

void LedBoxBase::SetDistance(int place, const VmPoint *ndist)
{
	win_pt.place = place;
	dist_set[0].x = ndist[0].x;
	dist_set[0].y = ndist[0].y;
	dist_set[1].x = ndist[1].x;
	dist_set[1].y = ndist[1].y;

	dist.x = dist_set[mode].x;
	dist.y = dist_set[mode].y;
}

void LedBoxBase::GetDistance(VmPoint *ndist)
{
	dist_set[mode].x = dist.x;
	dist_set[mode].y = dist.y;

	ndist[0].x = dist_set[0].x;
	ndist[0].y = dist_set[0].y;
	ndist[1].x = dist_set[1].x;
	ndist[1].y = dist_set[1].y;
}
