/** @file msgboard.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.26 -

	@brief [ message board ]
*/
#include "msgboard.h"
#include "common.h"
#include "emu.h"
#include "config.h"

#undef USE_BG_TRANSPARENT

MsgBoard::MsgBoard(HWND hWnd, EMU *pEmu) : DibSection()
{
	visible = true;
	inited = false;

	info.font = new CFont();
	msg.font = new CFont();

	hPen   = NULL;
	hBrush = NULL;

	// F
	fg = RGB(0x00, 0xc0, 0x80);
	bg = RGB(0x00, 0x40, 0x00);

	szWin.cx = 1;
	szWin.cy = 1;

	emu = pEmu;

	InitializeCriticalSection(&cs);
}

MsgBoard::~MsgBoard()
{
	if (enable) {
		if (hBrush)     DeleteObject(hBrush);
		if (hPen)       DeleteObject(hPen);
	}
	delete info.font;
	delete msg.font;

	DeleteCriticalSection(&cs);
}

/// ʏ
void MsgBoard::InitScreen(HDC hdc, int width, int height)
{
	// EBhETCY
	szWin.cx = width;
	szWin.cy = height;

	// Dib쐬
	CreateDib(hdc, width, 128);

	if (enable) {
		// `py̐ݒ
		hPen = CreatePen(PS_SOLID, 1, fg);
		SelectObject(hMainDib, hPen);

		// `puV̐ݒ
		hBrush = CreateSolidBrush(bg);
		SelectObject(hMainDib, hBrush);

		// gݒ
		SetRect(&msg.re, 0, 0, width, 63);
		// \ʒu
		msg.pt.x = 0; msg.pt.y = 0; info.place = 2;

		// gݒ
		SetRect(&info.re, 0, 64, width, 127);
		// \ʒu
		info.pt.x = 0; info.pt.y = 0; info.place = 1;

	}
	if (enable && SetFont()) {
		inited = true;
		msg.font->GetFontName(config.msgboard_msg_fontname);
		info.font->GetFontName(config.msgboard_info_fontname);
	}
	if (enable) {
		emu->out_log(LOG_INFO , _T("MsgBoard: OK"));
	} else {
		emu->out_log(LOG_ERROR, _T("MsgBoard: Failed"));
	}
}

/// tHgݒ
bool MsgBoard::SetFont()
{
	_TCHAR font_name[64];

	// tHgpX̐ݒ
	if (_tcslen(config.font_path) > 0 && !CFont::AddFontPath(config.font_path)) {
		emu->out_logf(LOG_WARN, _T("MsgBoard: Couldn't load font %s."), config.font_path);
	}

	msg.font->SetFont((HWND)NULL, config.msgboard_msg_fontname, config.msgboard_msg_fontsize, FW_NORMAL, fg);
	if (msg.font->GetFont() != NULL) {
		msg.font->GetFontName(font_name);
		emu->out_logf(LOG_INFO, _T("MsgBoard: Use \"%s\" for message."), font_name);
	} else {
		emu->out_logf(LOG_WARN, _T("MsgBoard: Couldn't load font \"%s\" for message."), config.msgboard_msg_fontname);
	}
	info.font->SetFont((HWND)NULL, config.msgboard_info_fontname, config.msgboard_info_fontsize, FW_BOLD, fg);
	if (info.font->GetFont() != NULL) {
		info.font->GetFontName(font_name);
		emu->out_logf(LOG_INFO, _T("MsgBoard: Use \"%s\" for info."), font_name);
	} else {
		emu->out_logf(LOG_WARN, _T("MsgBoard: Couldn't load font \"%s\" for info."), config.msgboard_info_fontname);
	}
	if (msg.font->GetFont() == NULL || info.font->GetFont() == NULL) {
		enable = false;
	} else {
		enable = true;
	}
	return enable;
}

/// o
void MsgBoard::draw(HDC hdc, msg_data_t &data)
{
	POINT pt;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// ʒǔvZ
		pt = data.pt;
		if (data.place & 1) {
			pt.x += szWin.cx - data.sz.cx;
		}
		if (data.place & 2) {
			pt.y += szWin.cy - data.sz.cy;
		}

		EnterCriticalSection(&cs);

		// CReLXgɃbZ[WRs[
#ifdef USE_BG_TRANSPARENT
		draw_text(hdc, data, pt.x, pt.y);
#else
		BitBlt(hdc, pt.x, pt.y,
			data.sz.cx, data.sz.cy,
			hMainDib, data.re.left, data.re.top, SRCCOPY);
#endif

		LeaveCriticalSection(&cs);
	}
}
void MsgBoard::draw(LPDIRECT3DSURFACE9 suf, msg_data_t &data)
{
	RECT  reDst;

	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// ʒǔvZ
		if (data.place & 1) {
			// ʉE_
			reDst.left = szWin.cx + data.pt.x - data.sz.cx;
			reDst.right = szWin.cx + data.pt.x;
		} else {
			// ʍ_
			reDst.left = data.pt.x;
			reDst.right = data.pt.x + data.sz.cx;
		}
		if (data.place & 2) {
			// ʉ_
			reDst.top = szWin.cy + data.pt.y - data.sz.cy;
			reDst.bottom = szWin.cy + data.pt.y;
		} else {
			// ʏオ_
			reDst.top = data.pt.y;
			reDst.bottom = data.pt.y + data.sz.cy;
		}

		EnterCriticalSection(&cs);

		// CReLXgɃbZ[WRs[
#ifdef USE_DIRECT3DX
		RECT  reSrc;
		SetRect(&reSrc, data.re.left, data.re.top, data.re.left + data.sz.cx, data.re.top + data.sz.cy);

		D3DXLoadSurfaceFromMemory(suf,
			NULL,
			&reDst,
			pMainBuf,
			D3DFMT_X8R8G8B8,
			szWin.cx * 4,
			NULL,
			&reSrc,
			D3DX_FILTER_NONE,
			0);
#else
		if (suf) {
			HDC hdc;
			HRESULT hre = suf->GetDC(&hdc);
			if (hre == D3D_OK) {
#ifdef USE_BG_TRANSPARENT
				draw_text(hdc, data, reDst.left, reDst.top);
#else
				BitBlt(hdc, reDst.left, reDst.top,
					data.sz.cx, data.sz.cy,
					hMainDib, data.re.left, data.re.top, SRCCOPY);
#endif
			}
			suf->ReleaseDC(hdc);
		}
#endif

		LeaveCriticalSection(&cs);
	}
}

void MsgBoard::Draw(HDC hdc)
{
	if (!visible || !inited || !enable) return;

	draw(hdc, msg);
	draw(hdc, info);
}

void MsgBoard::Draw(LPDIRECT3DSURFACE9 suf)
{
	if (!visible || !inited || !enable) return;

	draw(suf, msg);
	draw(suf, info);
}

/// obNobt@ɕ`
void MsgBoard::draw_text(msg_data_t &data)
{
	SIZE sz;
	int len = 0;

	if (enable && !data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		EnterCriticalSection(&cs);

		// tHgݒ
		SelectObject(hMainDib, data.font->GetFont());

		// 
		len = (int)_tcslen(it->msg);

#ifndef USE_BG_TRANSPARENT
		// wiF
		SetBkColor(hMainDib, bg);
		// F
		SetTextColor(hMainDib, data.font->GetFontColor());

		// g`
		FillRect(hMainDib, &data.re, hBrush);

		// `
		TextOut(hMainDib, data.re.left + 2, data.re.top + 2, it->msg, len);
#endif

		// ̘̕gvZ
		GetTextExtentPoint32(hMainDib, it->msg, len, &sz);

		LeaveCriticalSection(&cs);

		data.sz.cx = sz.cx + 4;
		data.sz.cy = sz.cy + 4;
	}
}

/// obNobt@ɕ`
void MsgBoard::draw_text(HDC hdc, msg_data_t &data, int left, int top)
{
	int len = 0;

	if (enable && !data.lists.empty()) {
		list_t::iterator it = data.lists.begin();

		// wiF
//		SetBkColor(hdc, bg);
		// wiF͓
		SetBkMode(hdc, TRANSPARENT);
		// F
		SetTextColor(hdc, data.font->GetFontColor());

		// g`
//		FillRect(hdc, &data.re, hBrush);

		// tHgݒ
		SelectObject(hdc, data.font->GetFont());

		// `
		len = (int)_tcslen(it->msg);
		TextOut(hdc, left + data.re.left + 2, top + 2, it->msg, len);
	}
}

/// bZ[Wݒ
void MsgBoard::Set(msg_data_t &data, const _TCHAR *str, int sec)
{
	item_t itm;

	EnterCriticalSection(&cs);

	_tcsncpy(itm.msg, str, 1023);
	itm.cnt = (60 * sec);

	data.lists.push_front(itm);
	if (data.lists.size() > 10) {
		data.lists.pop_back();
	}

	LeaveCriticalSection(&cs);

	if (!inited || !enable) return;

	draw_text(data);
}
void MsgBoard::SetMessage(const _TCHAR *str, int sec)
{
	Set(msg, str, sec);
}
void MsgBoard::SetInfo(const _TCHAR *str, int sec)
{
	Set(info, str, sec);
}
void MsgBoard::SetMessageF(const _TCHAR *format, ...)
{
	_TCHAR buf[1024];

	va_list ap;

	va_start(ap, format);
	_vstprintf(buf, format, ap);
	va_end(ap);

	SetMessage(buf);
}
void MsgBoard::SetInfoF(const _TCHAR *format, ...)
{
	_TCHAR buf[1024];

	va_list ap;

	va_start(ap, format);
	_vstprintf(buf, format, ap);
	va_end(ap);

	SetInfo(buf);
}

/// EBhETCYݒ
void MsgBoard::SetSize(int width, int height)
{
	szWin.cx = width;
	szWin.cy = height;
}

/// bZ[W`ʒuݒ
/// @param[in] cx X
/// @param[in] cy Y
/// @param[in] place 1:ʉE 2:ʉ
void MsgBoard::SetMessagePos(int cx, int cy, int place)
{
	msg.pt.x = cx;
	msg.pt.y = cy;
	msg.place = place;
}
/// `ʒuݒ
/// @param[in] cx X
/// @param[in] cy Y
/// @param[in] place 1:ʉE 2:ʉ
void MsgBoard::SetInfoPos(int cx, int cy, int place)
{
	info.pt.x = cx;
	info.pt.y = cy;
	info.place = place;
}

/// bZ[W폜
void MsgBoard::Delete(msg_data_t &data, const _TCHAR *str)
{
	bool redraw = false;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();
		list_t::iterator it_next;
		while(it != data.lists.end())	{
			it_next = it;
			it_next++;

			if (_tcscmp(str, it->msg) == 0) {
				data.lists.erase(it);
				redraw = true;
			}
			it = it_next;
		}
	}
	if (redraw) {
		draw_text(data);
	}
}
void MsgBoard::DeleteMessage(const _TCHAR *str)
{
	Delete(msg, str);
}
void MsgBoard::DeleteInfo(const _TCHAR *str)
{
	Delete(info, str);
}

/// JEg_E
void MsgBoard::count_down(msg_data_t &data)
{
	bool redraw = false;
	if (!data.lists.empty()) {
		list_t::iterator it = data.lists.begin();
		list_t::iterator it_next;
		while(it != data.lists.end())	{
			it_next = it;
			it_next++;

			if (it->cnt > 0) it->cnt--;
			if (it->cnt == 0) {
				data.lists.erase(it);
				redraw = true;
			}
			it = it_next;
		}
	}
	if (redraw) {
		draw_text(data);
	}
}
void MsgBoard::CountDown(void)
{
	count_down(msg);
	count_down(info);
}
