/** @file ffm_rec_video.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ record video using ffmpeg ]
*/

#include "ffm_rec_video.h"

#if defined(USE_REC_VIDEO) && defined(USE_REC_VIDEO_FFMPEG)

//extern "C" {
//#include <libavutil/opt.h>
//#include <libavutil/imgutils.h>
//};
#include "../rec_video.h"
#include "../../emu.h"

#ifndef _WIN32
#include <dlfcn.h>
#endif

//#ifdef _MSC_VER
//#pragma comment(lib, "avcodec.lib")
//#pragma comment(lib, "avutil.lib")
//#pragma comment(lib, "swscale.lib")
//#pragma comment(lib, "avformat.lib")
//#endif


// entry point of avcodec-56.dll
void (*f_avcodec_register_all)(void) = NULL;
AVCodec *(*f_avcodec_find_encoder)(enum AVCodecID id) = NULL;
int (*f_avcodec_open2)(AVCodecContext *avctx, const AVCodec *codec, AVDictionary **options) = NULL;
int (*f_avcodec_encode_video2)(AVCodecContext *avctx, AVPacket *avpkt,
                          const AVFrame *frame, int *got_packet_ptr) = NULL;
void (*f_av_init_packet)(AVPacket *pkt) = NULL;
void (*f_av_packet_rescale_ts)(AVPacket *pkt, AVRational tb_src, AVRational tb_dst) = NULL;
void (*f_av_free_packet)(AVPacket *pkt) = NULL;


// entry point of avformat-56.dll
void (*f_av_register_all)(void) = NULL;
AVFormatContext *(*f_avformat_alloc_context)(void) = NULL;
void (*f_avformat_free_context)(AVFormatContext *s) = NULL;
AVOutputFormat *(*f_av_guess_format)(const char *short_name,
                                const char *filename,
                                const char *mime_type) = NULL;
int (*f_avformat_query_codec)(const AVOutputFormat *ofmt, enum AVCodecID codec_id,
                         int std_compliance) = NULL;
AVStream *(*f_avformat_new_stream)(AVFormatContext *s, const AVCodec *c) = NULL;
int (*f_avformat_write_header)(AVFormatContext *s, AVDictionary **options) = NULL;
int (*f_av_write_frame)(AVFormatContext *s, AVPacket *pkt) = NULL;
int (*f_av_write_trailer)(AVFormatContext *s) = NULL;
int (*f_avio_open)(AVIOContext **s, const char *url, int flags) = NULL;
int64_t (*f_avio_size)(AVIOContext *s) = NULL;
int (*f_avio_close)(AVIOContext *s) = NULL;

// entry point of swscale-3.dll
struct SwsContext *(*f_sws_getContext)(int srcW, int srcH, enum AVPixelFormat srcFormat,
				int dstW, int dstH, enum AVPixelFormat dstFormat,
				int flags, SwsFilter *srcFilter,
				SwsFilter *dstFilter, const double *param) = NULL;
int (*f_sws_scale)(struct SwsContext *c, const uint8_t *const srcSlice[],
				const int srcStride[], int srcSliceY, int srcSliceH,
				uint8_t *const dst[], const int dstStride[]) = NULL;
void (*f_sws_freeContext)(struct SwsContext *swsContext) = NULL;

// entry point of avutil-54.dll
AVFrame *(*f_av_frame_alloc)(void) = NULL;
void (*f_av_frame_free)(AVFrame **frame) = NULL;
int (*f_av_image_alloc)(uint8_t *pointers[4], int linesizes[4],
                   int w, int h, enum AVPixelFormat pix_fmt, int align) = NULL;
void (*f_av_freep)(void *ptr) = NULL;
void (*f_av_log_set_callback)(void (*callback)(void*, int, const char*, va_list)) = NULL;
const char* (*f_av_default_item_name)(void* ctx) = NULL;
void (*f_av_log_set_level)(int level) = NULL;


void cb_log(void *avcl, int level, const char *fmt, va_list vl)
{
	extern EMU *emu;
	if (f_av_default_item_name && avcl) {
		const char *name = f_av_default_item_name(avcl);
		if (name) emu->out_logf(LOG_DEBUG, _T("ffmpeg: %s level:%d"), name, level);
	}
	emu->out_logv(LOG_DEBUG, fmt, vl);
}


FFM_REC_VIDEO::FFM_REC_VIDEO(EMU *new_emu, REC_VIDEO *new_vid)
{
	emu = new_emu;
	vid = new_vid;
	rec_fps = 0;
	memset(&rec_rect, 0, sizeof(rec_rect));
	rec_surface = NULL;
	rec_path = NULL;

	avio = NULL;
	fmtcont = NULL;

	codcont = NULL;
	frame = NULL;
	srcframe = NULL;
	sws_cont = NULL;

	library_loaded = false;

	// for load dll library
	hAVCodec = NULL;
	hAVFormat = NULL;
	hSWScale = NULL;
	hAVUtil = NULL;
}

FFM_REC_VIDEO::~FFM_REC_VIDEO()
{
	Release();
	Unload();
}

void FFM_REC_VIDEO::Release()
{
	if (fmtcont) {
		// finish and flush
		if (vid->NowRecording()) f_av_write_trailer(fmtcont);
		f_avformat_free_context(fmtcont);
		fmtcont = NULL;
	}
	if (avio) {
		f_avio_close(avio);
		avio = NULL;
	}
	if (sws_cont) {
		f_sws_freeContext(sws_cont);
		sws_cont = NULL;
	}
	if (srcframe) {
		f_av_frame_free(&srcframe);
		srcframe = NULL;
	}
	if (frame) {
		f_av_freep(&frame->data[0]);
		f_av_frame_free(&frame);
		frame = NULL;
	}
	if (codcont) {
		codcont = NULL;
	}
}

void FFM_REC_VIDEO::RemoveFile()
{
#if defined(_WIN32)
	DeleteFile(rec_path);
#else
	unlink(rec_path);
#endif
}

#if defined(_WIN32)
#define LOAD_LIB(handle, lib_base, version) { \
	_TCHAR libname[64]; \
	_tcscpy(libname, _T(lib_base)); \
	if (version) { \
		_tcscat(libname, _T("-")); \
		_tcscat(libname, version); \
	} \
	_tcscat(libname, _T(".dll")); \
	handle = LoadLibrary(libname); \
	if (!handle) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot load %s."), libname); \
		continue; \
	} else { \
		emu->out_logf(LOG_DEBUG, _T("Loaded %s."), libname); \
	} \
}
#define GET_ADDR(func, func_type, handle, func_name) { \
	func = (func_type) GetProcAddress(handle, func_name); \
	if (!func) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot get address of %s."), _T(func_name)); \
		continue; \
	} \
}
#define GET_ADDR_OPTIONAL(func, func_type, handle, func_name) { \
	func = (func_type) GetProcAddress(handle, func_name); \
	if (!func) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot get address of %s."), _T(func_name)); \
	} \
}
#define UNLOAD_LIB(handle) { \
	if (handle) FreeLibrary(handle); \
	handle = NULL; \
}
#else /* !_WIN32 */
#if defined(__APPLE__) && defined(__MACH__)
#define LOAD_LIB(handle, lib_base, version) { \
	char libname[64]; \
	strcpy(libname, "lib"); \
	strcat(libname, lib_base); \
	if (version) { \
		strcat(libname, "."); \
		strcat(libname, version); \
	} \
	strcat(libname, ".dylib"); \
	handle = dlopen(libname, RTLD_NOW | RTLD_GLOBAL); \
	if (!handle) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot load %s."), libname); \
		continue; \
	} else { \
		emu->out_logf(LOG_DEBUG, _T("Loaded %s."), libname); \
	} \
}
#else
#define LOAD_LIB(handle, lib_base, version) { \
	char libname[64]; \
	strcpy(libname, "lib"); \
	strcat(libname, lib_base); \
	strcat(libname, ".so"); \
	if (version) { \
		strcat(libname, "."); \
		strcat(libname, version); \
	} \
	handle = dlopen(libname, RTLD_NOW | RTLD_GLOBAL); \
	if (!handle) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot load %s."), libname); \
		continue; \
	} else { \
		emu->out_logf(LOG_DEBUG, _T("Loaded %s."), libname); \
	} \
}
#endif
#define GET_ADDR(func, func_type, handle, func_name) { \
	func = (func_type)dlsym(handle, func_name); \
	if (!func) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot get address of %s."), func_name); \
		continue; \
	} \
}
#define GET_ADDR_OPTIONAL(func, func_type, handle, func_name) { \
	func = (func_type)dlsym(handle, func_name); \
	if (!func) { \
		emu->out_logf(LOG_DEBUG, _T("Cannot get address of %s."), func_name); \
	} \
}
#define UNLOAD_LIB(handle) { \
	if (handle) dlclose(handle); \
	handle = NULL; \
}
#endif

static const struct {
	bool          end;
	const _TCHAR *avutil;
	const _TCHAR *swresample;
	const _TCHAR *avcodec;
	const _TCHAR *avformat;
	const _TCHAR *swscale;
} dllvers[] = {
	{ false, _T("54"),_T("1"),_T("56"),_T("56"),_T("3") },
	{ false, _T("55"),_T("2"),_T("57"),_T("57"),_T("4") },
	{ false, NULL, NULL, NULL, NULL, NULL },
	{ true,  NULL, NULL, NULL, NULL, NULL },
};

bool FFM_REC_VIDEO::Load()
{
	if (library_loaded) return true;

	// load dynamic link library file
	Unload();

	for(int i=0; dllvers[i].end != true; i++) {
		// Set entry points of AVUtil function
		LOAD_LIB(hAVUtil, "avutil", dllvers[i].avutil);
		GET_ADDR(f_av_frame_alloc, AVFrame *(*)(void), hAVUtil, "av_frame_alloc");
		GET_ADDR(f_av_frame_free, void (*)(AVFrame **frame), hAVUtil, "av_frame_free");
		GET_ADDR(f_av_image_alloc, int (*)(uint8_t *[4], int [4], int, int, enum AVPixelFormat, int), hAVUtil, "av_image_alloc");
		GET_ADDR(f_av_freep, void (*)(void *ptr), hAVUtil, "av_freep");
		GET_ADDR(f_av_log_set_callback, void (*)(void (*callback)(void*, int, const char*, va_list)), hAVUtil, "av_log_set_callback");
		GET_ADDR_OPTIONAL(f_av_default_item_name, const char* (*)(void* ctx), hAVUtil, "av_default_item_name");
		GET_ADDR_OPTIONAL(f_av_log_set_level, void (*)(int level), hAVUtil, "av_log_set_level");

		// Set entry points of SWResample function (reference from avcodec)
		LOAD_LIB(hSWScale, "swresample", dllvers[i].swresample);

		// Set entry points of AVCodec function
		LOAD_LIB(hAVCodec, "avcodec", dllvers[i].avcodec);
		GET_ADDR(f_avcodec_register_all, void (*)(void), hAVCodec, "avcodec_register_all");
		GET_ADDR(f_avcodec_find_encoder, AVCodec *(*)(enum AVCodecID id), hAVCodec, "avcodec_find_encoder");
		GET_ADDR(f_avcodec_open2, int (*)(AVCodecContext *avctx, const AVCodec *codec, AVDictionary **options), hAVCodec, "avcodec_open2");
		GET_ADDR(f_avcodec_encode_video2, int (*)(AVCodecContext *, AVPacket *, const AVFrame *, int *), hAVCodec, "avcodec_encode_video2");
		GET_ADDR(f_av_init_packet, void (*)(AVPacket *pkt), hAVCodec, "av_init_packet");
		GET_ADDR(f_av_packet_rescale_ts, void (*)(AVPacket *pkt, AVRational tb_src, AVRational tb_dst), hAVCodec, "av_packet_rescale_ts");
		GET_ADDR(f_av_free_packet, void (*)(AVPacket *pkt), hAVCodec, "av_free_packet");

		// Set entry points of AVFormat function
		LOAD_LIB(hAVFormat, "avformat", dllvers[i].avformat);
		GET_ADDR(f_av_register_all, void (*)(void), hAVFormat, "av_register_all");
		GET_ADDR(f_avformat_alloc_context, AVFormatContext *(*)(void), hAVFormat, "avformat_alloc_context");
		GET_ADDR(f_avformat_free_context, void (*)(AVFormatContext *s), hAVFormat, "avformat_free_context");
		GET_ADDR(f_av_guess_format, AVOutputFormat *(*)(const char *, const char *, const char *), hAVFormat, "av_guess_format");
		GET_ADDR(f_avformat_query_codec, int (*)(const AVOutputFormat *, enum AVCodecID, int), hAVFormat, "avformat_query_codec");
		GET_ADDR(f_avformat_new_stream, AVStream *(*)(AVFormatContext *s, const AVCodec *c), hAVFormat, "avformat_new_stream");
		GET_ADDR(f_avformat_write_header, int (*)(AVFormatContext *s, AVDictionary **options), hAVFormat, "avformat_write_header");
		GET_ADDR(f_av_write_frame, int (*)(AVFormatContext *s, AVPacket *pkt), hAVFormat, "av_write_frame");
		GET_ADDR(f_av_write_trailer, int (*)(AVFormatContext *s), hAVFormat, "av_write_trailer");
		GET_ADDR(f_avio_open, int (*)(AVIOContext **s, const char *url, int flags), hAVFormat, "avio_open");
		GET_ADDR(f_avio_size, int64_t (*)(AVIOContext *s), hAVFormat, "avio_size");
		GET_ADDR(f_avio_close, int (*)(AVIOContext *s), hAVFormat, "avio_close");

		// Set entry points of SWScale function
		LOAD_LIB(hSWScale, "swscale", dllvers[i].swscale);
		GET_ADDR(f_sws_getContext, struct SwsContext *(*)(int srcW, int srcH, enum AVPixelFormat srcFormat,
												   int dstW, int dstH, enum AVPixelFormat dstFormat,
												   int flags, SwsFilter *srcFilter,
												   SwsFilter *dstFilter, const double *param), hSWScale, "sws_getContext");
		GET_ADDR(f_sws_scale, int (*)(struct SwsContext *c, const uint8_t *const srcSlice[],
							   const int srcStride[], int srcSliceY, int srcSliceH,
							   uint8_t *const dst[], const int dstStride[]), hSWScale, "sws_scale");
		GET_ADDR(f_sws_freeContext, void (*)(struct SwsContext *swsContext), hSWScale, "sws_freeContext");

		library_loaded = true;
		break;
	}
	if (!library_loaded) goto FIN;

	// register all the codecs
	f_av_register_all();
    f_avcodec_register_all();

#ifdef _DEBUG
	f_av_log_set_callback(cb_log);
#else
	if (f_av_log_set_level) f_av_log_set_level(AV_LOG_INFO);
#endif
	return true;
FIN:
//	emu->out_log(LOG_ERROR, _("Couldn't load FFmpeg library."));
	return false;
}

void FFM_REC_VIDEO::Unload()
{
	UNLOAD_LIB(hAVCodec);
	UNLOAD_LIB(hAVFormat);
	UNLOAD_LIB(hSWScale);
	UNLOAD_LIB(hAVUtil);
	library_loaded = false;
}

bool FFM_REC_VIDEO::IsEnabled()
{
	return Load();
}

const _TCHAR **FFM_REC_VIDEO::GetCodecList()
{
	static const _TCHAR *list[] = {
//		_("FFmpeg video"),
//		_("H.263"),
		_("H.264"),
		_("MPEG-2"),
		_("MPEG-4"),
		_("Motion JPEG"),
//		_("MS Video-1"),
		_("WMV1"),
		_("WMV2"),
		NULL };
	return list;
}

const _TCHAR **FFM_REC_VIDEO::GetQualityList()
{
	static const _TCHAR *list[] = {
		_("Max (50Mbps)"),
		_("High (5Mbps)"),
		_("Normal (500Kbps)"),
		_("Low (50Kbps)"),
		_("Min (5Kbps)"),
		NULL };
	return list;
}

bool FFM_REC_VIDEO::Start(_TCHAR *path, int fps, VmRectWH *srcrect, DibSection *srcsurface, bool show_dialog)
{
	const AVCodecID codec_ids[] = {
//		AV_CODEC_ID_FFV1, // mac ng
//		AV_CODEC_ID_H263, // mac ng
		AV_CODEC_ID_H264, // win ok / mac need libx264
		AV_CODEC_ID_MPEG2VIDEO, // win ok / mac ok
		AV_CODEC_ID_MPEG4, // win ok / mac warn quick time cannot play it.
		AV_CODEC_ID_MJPEG,
//		AV_CODEC_ID_MSVIDEO1, //
		AV_CODEC_ID_WMV1, // win ok
		AV_CODEC_ID_WMV2, // win ok
		AV_CODEC_ID_NONE
//		AV_CODEC_ID_MPEG1VIDEO, // win ng
//		AV_CODEC_ID_RAWVIDEO, // win ok
	};
	const int bit_rates[] = {
		 50000000,	// max
	 	  5000000,	// high
		   500000,	// normal
		    50000,	// low
		     5000,	// min
		0
	};

	AVCodecID codec_id = codec_ids[emu->get_parami(VM::ParamRecVideoCodec)];
	AVCodec *codec = NULL;
	AVStream *stream = NULL;
	char name[64];
	int ret;

	if (!Load()) return false;

	name[0] = '\0';

	if (path != NULL) {
		// format name
		switch(codec_id) {
		case AV_CODEC_ID_FFV1:
			strcat(name, "ffm");
			break;
		case AV_CODEC_ID_H263:
			strcat(name, "3gp");
			break;
		case AV_CODEC_ID_H264:
			strcat(name, "mp4");
			break;
		case AV_CODEC_ID_MPEG4:
			strcat(name, "mp4");
//			strcat(name, "mpeg4");
			break;
		case AV_CODEC_ID_MPEG2VIDEO:
			strcat(name, "vob");
			break;
		case AV_CODEC_ID_WMV1:
			strcat(name, "asf");
			break;
		case AV_CODEC_ID_WMV2:
			strcat(name, "wmv2");
			break;
		case AV_CODEC_ID_MJPEG:
			strcat(name, "mjpeg");
			break;
		case AV_CODEC_ID_MSVIDEO1:
			strcat(name, "avi");
			break;
		default:
			break;
		}
		// extention
		switch(codec_id) {
		case AV_CODEC_ID_FFV1:
			_tcscat(path, _T(".ffm"));
			break;
		case AV_CODEC_ID_H263:
			_tcscat(path, _T(".3gp"));
			break;
		case AV_CODEC_ID_H264:
		case AV_CODEC_ID_MPEG4:
			_tcscat(path, _T(".mp4"));
			break;
		case AV_CODEC_ID_MPEG2VIDEO:
			_tcscat(path, _T(".mpg"));
			break;
		case AV_CODEC_ID_WMV1:
		case AV_CODEC_ID_WMV2:
			_tcscat(path, _T(".wmv"));
			break;
		case AV_CODEC_ID_MSVIDEO1:
			_tcscat(path, _T(".avi"));
			break;
		default:
#if (defined(__APPLE__) && defined(__MACH__))
			_tcscat(path, _T(".mov"));	// quicktime
#else
			_tcscat(path, _T(".avi"));	// avi
#endif
			break;
		}
		rec_path = path;
	}
	rec_fps = fps;
	if (srcrect != NULL) {
		rec_rect = *srcrect;
	}
	if (srcsurface != NULL) {
		rec_surface = srcsurface;
	}
	if (rec_fps <= 0 || rec_rect.w <= 0 || rec_rect.h <= 0 || rec_surface == NULL) {
		return false;
	}

	emu->out_logf(LOG_DEBUG, _T("FFM_REC_VIDEO::Start: %d %s"), codec_id, name);

	// open file for output
	char crec_path[_MAX_PATH];
	UTILITY::wcs_to_mbs(crec_path, rec_path);
	ret = f_avio_open(&avio, crec_path, AVIO_FLAG_WRITE);
	if (ret < 0) {
		emu->out_logf(LOG_ERROR, _T("avio_open failed: %d"), ret);
		goto FIN;
	}
	// alloc format context
	fmtcont = f_avformat_alloc_context();
	if (!fmtcont) {
		emu->out_log(LOG_ERROR, _T("avformat_alloc_context failed."));
		goto FIN;
	}
	// attach io context with format context
	fmtcont->pb = avio;

	// guess format by name
	fmtcont->oformat = f_av_guess_format(name, crec_path, NULL);
	if (!fmtcont->oformat) {
		emu->out_log(LOG_ERROR, _T("av_guess_format failed."));
		goto FIN;
	}
	// whether use this codec on specified file format
	ret = f_avformat_query_codec(fmtcont->oformat, codec_id, FF_COMPLIANCE_NORMAL);
	if (ret <= 0) {
		emu->out_logf(LOG_ERROR, _T("avformat_query_codec failed: %d"), ret);
		goto FIN;
	}
	strncpy(fmtcont->filename, crec_path, sizeof(fmtcont->filename));
	fmtcont->oformat->audio_codec = AV_CODEC_ID_NONE;
	fmtcont->oformat->video_codec = codec_id;
	fmtcont->oformat->subtitle_codec = AV_CODEC_ID_NONE;

	// find the specified video encoder
	codec = f_avcodec_find_encoder(codec_id);
	if (!codec) {
		emu->out_logf(LOG_ERROR, _T("avcodec_find_encoder: not found codec id: %d"), codec_id);
		goto FIN;
	}

	// create a stream in the output file container
	stream = f_avformat_new_stream(fmtcont, codec);
	if (!stream) {
		emu->out_log(LOG_ERROR, _T("avformat_new_stream failed."));
		goto FIN;
	}
	codcont = stream->codec;

	codcont->width = rec_rect.w;
	codcont->height = rec_rect.h;

	// frames per second
	codcont->time_base.num = 1;
	codcont->time_base.den = rec_fps;
	stream->time_base = codcont->time_base;

	codcont->gop_size = rec_fps;
	codcont->max_b_frames = 1;
	codcont->pix_fmt = AV_PIX_FMT_YUV420P;

	// put bit rate
	codcont->bit_rate = bit_rates[emu->get_parami(VM::ParamRecVideoQuality)];

	if (fmtcont->oformat->flags & AVFMT_GLOBALHEADER) {
		codcont->flags |= CODEC_FLAG_GLOBAL_HEADER;
	}
	switch(codec_id) {
	case AV_CODEC_ID_H264:
//		av_opt_set(codcont->priv_data, "preset", "slow", 0);
		break;
	case AV_CODEC_ID_H263:
	case AV_CODEC_ID_WMV1:
	case AV_CODEC_ID_WMV2:
		codcont->max_b_frames = 0;
		break;
	case AV_CODEC_ID_MJPEG:
		codcont->pix_fmt = AV_PIX_FMT_YUVJ444P;
		codcont->max_b_frames = 0;
		break;
	case AV_CODEC_ID_MSVIDEO1:
			codcont->pix_fmt = AV_PIX_FMT_RGB555;
			break;
	case AV_CODEC_ID_RAWVIDEO:
#if defined(_RGB888)
		codcont->pix_fmt = AV_PIX_FMT_RGB32;
#elif defined(_RGB565)
		codcont->pix_fmt = AV_PIX_FMT_RGB565;
#elif defined(_RGB555)
		codcont->pix_fmt = AV_PIX_FMT_RGB555;
#endif
		break;
	default:
		break;
	}

	// open it
	ret = f_avcodec_open2(codcont, codec, NULL);
	if (ret < 0) {
		emu->out_logf(LOG_ERROR, _T("avcodec_open2 failed: %d"), ret);
		goto FIN;
	}

	// create frame
	frame = f_av_frame_alloc();
	if (!frame) {
		emu->out_log(LOG_ERROR, _T("av_frame_alloc failed."));
		goto FIN;
	}
	frame->format = codcont->pix_fmt;
	frame->width  = codcont->width;
	frame->height = codcont->height;
	frame->pts = 0;

//	frame->quality = 1; // FF_LAMBDA_MAX;

	// allocate data buffer
	ret = f_av_image_alloc(frame->data, frame->linesize, frame->width, frame->height, (AVPixelFormat)frame->format, 32);
	if (ret < 0) {
		emu->out_log(LOG_ERROR, _T("av_image_alloc failed."));
		goto FIN;
	}

	// create frame for source image
	srcframe = f_av_frame_alloc();
	if (!srcframe) {
		emu->out_log(LOG_ERROR, _T("av_frame_alloc failed."));
		goto FIN;
	}
	srcframe->width = frame->width;
	srcframe->height = frame->height;
#if defined(_RGB888)
#if defined(__APPLE__) && defined(__MACH__)
	srcframe->format = AV_PIX_FMT_0BGR32;
#else
	srcframe->format = AV_PIX_FMT_0RGB32;
#endif
	srcframe->linesize[0] = frame->width * 4;
#elif defined(_RGB565)
	srcframe->format = AV_PIX_FMT_RGB565;
	srcframe->linesize[0] = frame->width * 2;
#elif defined(_RGB555)
	srcframe->format = AV_PIX_FMT_RGB555;
	srcframe->linesize[0] = frame->width * 2;
#endif

	// for convert frame image
	sws_cont = f_sws_getContext(srcframe->width, srcframe->height, (AVPixelFormat)srcframe->format,
	frame->width, frame->height, (AVPixelFormat)frame->format, SWS_FAST_BILINEAR, NULL, NULL, NULL);
	if (!sws_cont) {
		emu->out_log(LOG_ERROR, _T("sws_getContext failed."));
		goto FIN;
	}

	// write the header of the output file container
    ret = f_avformat_write_header(fmtcont, NULL);
	if (ret < 0) {
		emu->out_logf(LOG_ERROR, _T("avformat_write_header failed: %d"), ret);
		goto FIN;
	}

	return true;
FIN:
	Release();
	RemoveFile();
	emu->out_log(LOG_ERROR, (CT)_("Couldn't start recording video."));
	return false;
}

void FFM_REC_VIDEO::Stop()
{
	// get the delayed frames
	int ret;
	AVPacket pkt;
	f_av_init_packet(&pkt);
	pkt.data = NULL;    // packet data will be allocated by the encoder
	pkt.size = 0;
	int got_output = 1;
	while (got_output) {
		ret = f_avcodec_encode_video2(codcont, &pkt, NULL, &got_output);
		if (ret < 0) {
			emu->out_logf(LOG_ERROR, _T("Stop: avcodec_encode_video2 failed: %d"), ret);
			break;
		}
		if (got_output) {
			f_av_packet_rescale_ts(&pkt, codcont->time_base, fmtcont->streams[0]->time_base);
			f_av_write_frame(fmtcont, &pkt);
			f_av_free_packet(&pkt);
		}
	}

	Release();
}

bool FFM_REC_VIDEO::Restart()
{
	vid->Stop();
	return vid->Start(RECORD_VIDEO_TYPE_FFMPEG, -1, NULL, NULL, false);
}

bool FFM_REC_VIDEO::Record(DibSection *srcsurface)
{
	bool rc = true;

	int ret, got_output;
	AVPacket pkt;


	// convert frame
	srcframe->data[0] = (uint8_t *)rec_surface->GetBuffer();
    ret = f_sws_scale(sws_cont, (uint8_t const * const *)srcframe->data, srcframe->linesize, 0, srcframe->height,
		  frame->data, frame->linesize);

	// encode 1 frame of video
	f_av_init_packet(&pkt);
	pkt.data = NULL;    // packet data will be allocated by the encoder
	pkt.size = 0;

	frame->pts++;

	// encode the image
	ret = f_avcodec_encode_video2(codcont, &pkt, frame, &got_output);
	if (ret < 0) {
		emu->out_logf(LOG_ERROR,_T("Record: avcodec_encode_video2: %d"), ret);
		rc = false;
		vid->Stop();
	} else if (got_output) {
		f_av_packet_rescale_ts(&pkt, codcont->time_base, fmtcont->streams[0]->time_base);
		f_av_write_frame(fmtcont, &pkt);
		f_av_free_packet(&pkt);
#if defined(_M_X64) || defined(x86_64) || defined(__x86_64)
		// if file size > (2TB - 16MB), create new file
		if(f_avio_size(avio) >= 0x7fffffffff000000L)
#else
		// if file size > (2GB - 16MB), create new file
		if(f_avio_size(avio) >= 0x7f000000L)
#endif
		{
			vid->Stop();
			rc = vid->Start(RECORD_VIDEO_TYPE_FFMPEG, -1, NULL, NULL, false);
		}

	}
	return rc;
}

#endif /* USE_REC_VIDEO && USE_REC_VIDEO_FFMPEG */
