/** @file keyrecord.h

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.09.12

	@brief [ key recording ]
*/

#ifndef _KEYRECORD_H_
#define _KEYRECORD_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "../../config.h"
#include "../../fileio.h"

#ifdef _DEBUG
#define _DEBUG_KEYRECORD
#endif

/// for mouse
typedef struct st_mst {
		int pos;		///< x/y axis
		uint8 dir;		///< axis direction

		uint8 btn;		///< button left/right  b7:on/off b6:trigger on/off
		uint8 prev_btn;
} t_mouse_stat;


class KEYRECORD
{
private:
	EMU *emu;
	DEVICE *dev;

	int *mouse_stat;

	uint8 *key_scan_code_ptr;
	int *counter_ptr;
	int *remain_count_ptr;
	uint8 *kb_mode_ptr;

	// key record
	FILEIO *fkro, *fkri;
	uint8 vm_key_recp_stat[KEYBIND_KEYS];
	uint8 vm_key_recr_stat[KEYBIND_KEYS];
#ifdef USE_LIGHTPEN
	int   lpen_recp_stat[3];
	int   lpen_recr_stat[3];
#endif
	t_mouse_stat mouse_recp_stat[2];
	t_mouse_stat mouse_recr_stat[2];
	uint8 joypia_recp_stat[2];
	uint8 joypia_recr_stat[2];
	char  rec_key_tmp_buff[512];
	char  rec_key_rec_buff[512];
	int64 key_rec_sum_clock;
#ifdef _DEBUG
	uint8 vm_key_dbg_stat[KEYBIND_KEYS];
#endif

public:
	KEYRECORD(EMU* parent_emu);
	~KEYRECORD();

	void reading_keys(int num);
	bool recording_keys(int code, bool pressed);
	void recording_system_keys(int code, int pressed);
	void recording_mouse_status(t_mouse_stat *mst);
	void recording_joypia_status(uint8 *joy);
#ifdef USE_LIGHTPEN
	void recording_lightpen_status();
#endif

	bool play_reckey(const _TCHAR* filename);
	bool record_reckey(const _TCHAR* filename);
	void stop_reckey(bool stop_play = true, bool stop_record = true);

	void set_context(DEVICE* value) {
		dev = value;
	}
	void set_key_scan_code_ptr(uint8 *value) {
		key_scan_code_ptr = value;
	}
	void set_counter_ptr(int *value) {
		counter_ptr = value;
	}
	void set_remain_count_ptr(int *value) {
		remain_count_ptr = value;
	}
	void set_kb_mode_ptr(uint8 *value) {
		kb_mode_ptr = value;
	}
#ifdef USE_LIGHTPEN
	inline int get_lpen_recp_stat(int num) {
		return lpen_recp_stat[num];
	}
#endif
	inline t_mouse_stat *get_mouse_recp_stat() {
		return mouse_recp_stat;
	}
	inline uint8 get_vm_key_recp_stat(int code) {
		return vm_key_recp_stat[code];
	}
};

#endif
