/** @file timer.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ timer ]
*/

#ifndef _TIMER_H_
#define _TIMER_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "../../config.h"

//#define SIG_TIMER_VSYNC		1
#define SIG_TIMER_1MSEC		2

#define ADDR_TIMER_IRQ	0xffca
#define ADDR_TIME_MASK	0xffd4

/**
	@brief timer - process the timer irq signal
*/
class TIMER : public DEVICE
{
private:
	DEVICE *d_board;

	uint8	timer_irq;
	uint8	time_mask;

	uint8	cnt_2msec;

	int     register_id;

	//for resume
#pragma pack(1)
	struct vm_state_st {
		uint8	timer_irq;
		uint8	time_mask;
		uint8   cnt_2msec;
		char    reserved1;
		int     register_id;

		char  reserved[8];
	};
#pragma pack()

	void update_timer_clock();
	void update_fast_timer_clock();

public:
	TIMER(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("TIMER");
	}
	~TIMER() {}

	// common functions
	void initialize();
	void reset();
	void write_signal(int id, uint32 data, uint32 mask);
	void event_vline(int v, int clock);

	void write_io8(uint32 addr, uint32 data);
	uint32 read_io8(uint32 addr);

	// unique functions
	void set_context_board(DEVICE* device) {
		d_board = device;
	}

	void event_callback(int event_id, int err);

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);
};

#endif

