/** @file mc6809dasm.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.08.29 -

	@brief [ mc6809 disassembler ]
*/

#include "mc6809dasm.h"
#include "device.h"

#ifdef EMABLE_DISASSEMBLER

#undef OUTPUT_TRACELOG

struct opcode_st {
	uint8   c;
	_TCHAR *s;
};

static const opcode_st opcodes0[] =
{
{ 0x00, _T("NEG") },
{ 0x01, _T("??") },
{ 0x02, _T("??") },
{ 0x03, _T("COM") },
{ 0x04, _T("LSR") },
{ 0x05, _T("??") },
{ 0x06, _T("ROR") },
{ 0x07, _T("ASR") },
{ 0x08, _T("ASL") },
{ 0x09, _T("ROL") },
{ 0x0A, _T("DEC") },
{ 0x0B, _T("??") },
{ 0x0C, _T("INC") },
{ 0x0D, _T("TST") },
{ 0x0E, _T("JMP") },
{ 0x0F, _T("CLR") },

{ 0x10, _T("page1") },
{ 0x11, _T("page2") },
{ 0x12, _T("NOP") },
{ 0x13, _T("SYNC") },
{ 0x14, _T("??") },
{ 0x15, _T("??") },
{ 0x16, _T("LBRA") },
{ 0x17, _T("LBSR") },
{ 0x18, _T("??") },
{ 0x19, _T("DAA") },
{ 0x1A, _T("ORCC") },
{ 0x1B, _T("??") },
{ 0x1C, _T("ANDCC") },
{ 0x1D, _T("SEX") },
{ 0x1E, _T("EXG") },
{ 0x1F, _T("TFR") },

{ 0x20, _T("BRA") },
{ 0x21, _T("BRN") },
{ 0x22, _T("BHI") },
{ 0x23, _T("BLS") },
{ 0x24, _T("BCC") },
{ 0x25, _T("BCS") },
{ 0x26, _T("BNE") },
{ 0x27, _T("BEQ") },
{ 0x28, _T("BVC") },
{ 0x29, _T("BVS") },
{ 0x2A, _T("BPL") },
{ 0x2B, _T("BMI") },
{ 0x2C, _T("BGE") },
{ 0x2D, _T("BLT") },
{ 0x2E, _T("BGT") },
{ 0x2F, _T("BLE") },

{ 0x30, _T("LEAX") },
{ 0x31, _T("LEAY") },
{ 0x32, _T("LEAS") },
{ 0x33, _T("LEAU") },
{ 0x34, _T("PSHS") },
{ 0x35, _T("PULS") },
{ 0x36, _T("PSHU") },
{ 0x37, _T("PULU") },
{ 0x38, _T("??") },
{ 0x39, _T("RTS") },
{ 0x3A, _T("ABX") },
{ 0x3B, _T("RTI") },
{ 0x3C, _T("CWAI") },
{ 0x3D, _T("MUL") },
{ 0x3E, _T("??") },
{ 0x3F, _T("SWI") },

{ 0x40, _T("NEGA") },
{ 0x41, _T("??") },
{ 0x42, _T("??") },
{ 0x43, _T("COMA") },
{ 0x44, _T("LSRA") },
{ 0x45, _T("??") },
{ 0x46, _T("RORA") },
{ 0x47, _T("ASRA") },
{ 0x48, _T("ASLA") },
{ 0x49, _T("ROLA") },
{ 0x4A, _T("DECA") },
{ 0x4B, _T("??") },
{ 0x4C, _T("INCA") },
{ 0x4D, _T("TSTA") },
{ 0x4E, _T("??") },
{ 0x4F, _T("CLRA") },

{ 0x50, _T("NEGB") },
{ 0x51, _T("??") },
{ 0x52, _T("??") },
{ 0x53, _T("COMB") },
{ 0x54, _T("LSRB") },
{ 0x55, _T("??") },
{ 0x56, _T("RORB") },
{ 0x57, _T("ASRB") },
{ 0x58, _T("ASLB") },
{ 0x59, _T("ROLB") },
{ 0x5A, _T("DECB") },
{ 0x5B, _T("??") },
{ 0x5C, _T("INCB") },
{ 0x5D, _T("TSTB") },
{ 0x5E, _T("??") },
{ 0x5F, _T("CLRB") },

{ 0x60, _T("NEG") },
{ 0x61, _T("??") },
{ 0x62, _T("??") },
{ 0x63, _T("COM") },
{ 0x64, _T("LSR") },
{ 0x65, _T("??") },
{ 0x66, _T("ROR") },
{ 0x67, _T("ASR") },
{ 0x68, _T("ASL") },
{ 0x69, _T("ROL") },
{ 0x6A, _T("DEC") },
{ 0x6B, _T("??") },
{ 0x6C, _T("INC") },
{ 0x6D, _T("TST") },
{ 0x6E, _T("JMP") },
{ 0x6F, _T("CLR") },

{ 0x70, _T("NEG") },
{ 0x71, _T("??") },
{ 0x72, _T("??") },
{ 0x73, _T("COM") },
{ 0x74, _T("LSR") },
{ 0x75, _T("??") },
{ 0x76, _T("ROR") },
{ 0x77, _T("ASR") },
{ 0x78, _T("ASL") },
{ 0x79, _T("ROL") },
{ 0x7A, _T("DEC") },
{ 0x7B, _T("??") },
{ 0x7C, _T("INC") },
{ 0x7D, _T("TST") },
{ 0x7E, _T("JMP") },
{ 0x7F, _T("CLR") },

{ 0x80, _T("SUBA") },
{ 0x81, _T("CMPA") },
{ 0x82, _T("SBCA") },
{ 0x83, _T("SUBD") },
{ 0x84, _T("ANDA") },
{ 0x85, _T("BITA") },
{ 0x86, _T("LDA") },
{ 0x87, _T("STA??") },
{ 0x88, _T("EORA") },
{ 0x89, _T("ADCA") },
{ 0x8A, _T("ORA") },
{ 0x8B, _T("ADDA") },
{ 0x8C, _T("CMPX") },
{ 0x8D, _T("BSR") },
{ 0x8E, _T("LDX") },
{ 0x8F, _T("STX??") },

{ 0x90, _T("SUBA") },
{ 0x91, _T("CMPA") },
{ 0x92, _T("SBCA") },
{ 0x93, _T("SUBD") },
{ 0x94, _T("ANDA") },
{ 0x95, _T("BITA") },
{ 0x96, _T("LDA") },
{ 0x97, _T("STA") },
{ 0x98, _T("EORA") },
{ 0x99, _T("ADCA") },
{ 0x9A, _T("ORA") },
{ 0x9B, _T("ADDA") },
{ 0x9C, _T("CMPX") },
{ 0x9D, _T("JSR") },
{ 0x9E, _T("LDX") },
{ 0x9F, _T("STX") },

{ 0xA0, _T("SUBA") },
{ 0xA1, _T("CMPA") },
{ 0xA2, _T("SBCA") },
{ 0xA3, _T("SUBD") },
{ 0xA4, _T("ANDA") },
{ 0xA5, _T("BITA") },
{ 0xA6, _T("LDA") },
{ 0xA7, _T("STA") },
{ 0xA8, _T("EORA") },
{ 0xA9, _T("ADCA") },
{ 0xAA, _T("ORA") },
{ 0xAB, _T("ADDA") },
{ 0xAC, _T("CMPX") },
{ 0xAD, _T("JSR") },
{ 0xAE, _T("LDX") },
{ 0xAF, _T("STX") },

{ 0xB0, _T("SUBA") },
{ 0xB1, _T("CMPA") },
{ 0xB2, _T("SBCA") },
{ 0xB3, _T("SUBD") },
{ 0xB4, _T("ANDA") },
{ 0xB5, _T("BITA") },
{ 0xB6, _T("LDA") },
{ 0xB7, _T("STA") },
{ 0xB8, _T("EORA") },
{ 0xB9, _T("ADCA") },
{ 0xBA, _T("ORA") },
{ 0xBB, _T("ADDA") },
{ 0xBC, _T("CMPX") },
{ 0xBD, _T("JSR") },
{ 0xBE, _T("LDX") },
{ 0xBF, _T("STX") },

{ 0xC0, _T("SUBB") },
{ 0xC1, _T("CMPB") },
{ 0xC2, _T("SBCB") },
{ 0xC3, _T("ADDD") },
{ 0xC4, _T("ANDB") },
{ 0xC5, _T("BITB") },
{ 0xC6, _T("LDB") },
{ 0xC7, _T("STB??") },
{ 0xC8, _T("EORB") },
{ 0xC9, _T("ADCB") },
{ 0xCA, _T("ORB") },
{ 0xCB, _T("ADDB") },
{ 0xCC, _T("LDD") },
{ 0xCD, _T("STD??") },
{ 0xCE, _T("LDU") },
{ 0xCF, _T("STU??") },

{ 0xD0, _T("SUBB") },
{ 0xD1, _T("CMPB") },
{ 0xD2, _T("SBCB") },
{ 0xD3, _T("ADDD") },
{ 0xD4, _T("ANDB") },
{ 0xD5, _T("BITB") },
{ 0xD6, _T("LDB") },
{ 0xD7, _T("STB") },
{ 0xD8, _T("EORB") },
{ 0xD9, _T("ADCB") },
{ 0xDA, _T("ORB") },
{ 0xDB, _T("ADDB") },
{ 0xDC, _T("LDD") },
{ 0xDD, _T("STD") },
{ 0xDE, _T("LDU") },
{ 0xDF, _T("STU") },

{ 0xE0, _T("SUBB") },
{ 0xE1, _T("CMPB") },
{ 0xE2, _T("SBCB") },
{ 0xE3, _T("ADDD") },
{ 0xE4, _T("ANDB") },
{ 0xE5, _T("BITB") },
{ 0xE6, _T("LDB") },
{ 0xE7, _T("STB") },
{ 0xE8, _T("EORB") },
{ 0xE9, _T("ADCB") },
{ 0xEA, _T("ORB") },
{ 0xEB, _T("ADDB") },
{ 0xEC, _T("LDD") },
{ 0xED, _T("STD") },
{ 0xEE, _T("LDU") },
{ 0xEF, _T("STU") },

{ 0xF0, _T("SUBB") },
{ 0xF1, _T("CMPB") },
{ 0xF2, _T("SBCB") },
{ 0xF3, _T("ADDD") },
{ 0xF4, _T("ANDB") },
{ 0xF5, _T("BITB") },
{ 0xF6, _T("LDB") },
{ 0xF7, _T("STB") },
{ 0xF8, _T("EORB") },
{ 0xF9, _T("ADCB") },
{ 0xFA, _T("ORB") },
{ 0xFB, _T("ADDB") },
{ 0xFC, _T("LDD") },
{ 0xFD, _T("STD") },
{ 0xFE, _T("LDU") },
{ 0xFF, _T("STU") },
{ 0x00, NULL }
};

// ex1 opcodes (0x10 0x..)
static const opcode_st opcodes1[] =
{
{ 0x21, _T("LBRN") },
{ 0x22, _T("LBHI") },
{ 0x23, _T("LBLS") },
{ 0x24, _T("LBCC") },
{ 0x25, _T("LBCS") },
{ 0x26, _T("LBNE") },
{ 0x27, _T("LBEQ") },
{ 0x28, _T("LBVC") },
{ 0x29, _T("LBVS") },
{ 0x2A, _T("LBPL") },
{ 0x2B, _T("LBMI") },
{ 0x2C, _T("LBGE") },
{ 0x2D, _T("LBLT") },
{ 0x2E, _T("LBGT") },
{ 0x2F, _T("LBLE") },
{ 0x3F, _T("SWI2") },
{ 0x83, _T("CMPD") },
{ 0x8C, _T("CMPY") },
{ 0x8E, _T("LDY") },
{ 0x93, _T("CMPD") },
{ 0x9C, _T("CMPY") },
{ 0x9E, _T("LDY") },
{ 0x9F, _T("STY") },
{ 0xA3, _T("CMPD") },
{ 0xAC, _T("CMPY") },
{ 0xAE, _T("LDY") },
{ 0xAF, _T("STY") },
{ 0xB3, _T("CMPD") },
{ 0xBC, _T("CMPY") },
{ 0xBE, _T("LDY") },
{ 0xBF, _T("STY") },
{ 0xCE, _T("LDS") },
{ 0xDE, _T("LDS") },
{ 0xDF, _T("STS") },
{ 0xEE, _T("LDS") },
{ 0xEF, _T("STS") },
{ 0xFE, _T("LDS") },
{ 0xFF, _T("STS") },
{ 0x00, NULL }
};

// ex2 opcodes (0x11 0x..)
static const opcode_st opcodes2[] =
{
{ 0x3F, _T("SWI3") },
{ 0x83, _T("CMPU") },
{ 0x8C, _T("CMPS") },
{ 0x93, _T("CMPU") },
{ 0x9C, _T("CMPS") },
{ 0xA3, _T("CMPU") },
{ 0xAC, _T("CMPS") },
{ 0xB3, _T("CMPU") },
{ 0xBC, _T("CMPS") },
{ 0x00, NULL }
};

#endif /* ENABLE_DISASSEMBLER */

MC6809DASM::MC6809DASM()
{
#ifdef EMABLE_DISASSEMBLER
	phyaddr = 0;
	phyaddr_tmp = 0;
	inpc = 0;
	indp = 0;
	incc = 0;
	ina = 0;
	inb = 0;
	inx = 0;
	iny = 0;
	ins = 0;
	inu = 0;
	memset(opcode, 0, sizeof(opcode));
	memset(&memdata, 0, sizeof(memdata));
	codelen = 0;
	triggeron = -1;
	swi2fire = 0;
	sc_cat = sc_idx = 0;
	memset(pcstack, 0, sizeof(pcstack));
	pcstackpos = 0;
	loop_neg_count = MC6809DASM_PCSTACK_COUNT;
	loop_started = false;
	fp = NULL;

	now_halt = 0;
	now_reset = 0;
	now_nmi = 0;
	now_irq = 0;
	now_firq = 0;

	ti = 0;
	ti_prev = 0;

	cur_time = 0;
	cur_time_prev = 0;
	cur_time_sum = 0;
#endif
}

MC6809DASM::~MC6809DASM()
{
	close();
}

bool MC6809DASM::open()
{
#if defined(EMABLE_DISASSEMBLER) && defined(OUTPUT_TRACELOG)
	_TCHAR file[_MAX_PATH];

	time(&ti);
	tim = localtime(&ti);

	_stprintf(file, _T("trace%04d%02d%02d%02d%02d%02d.log")
		, tim->tm_year + 1900
		, tim->tm_mon + 1
		, tim->tm_mday
		, tim->tm_hour
		, tim->tm_min
		, tim->tm_sec
	);
	
	if (fp != NULL) close();
	fp = _tfopen(file, _T("w"));
	return (fp != NULL);
#else
	return false;
#endif
}

void MC6809DASM::close()
{
#ifdef EMABLE_DISASSEMBLER
	if (fp != NULL) fclose(fp);
	fp = NULL;
#endif
}

#ifdef EMABLE_DISASSEMBLER
void MC6809DASM::print_localtime()
{
	if (fp == NULL) return;
#ifndef _WIN32
	time(&ti);
	if (ti == ti_prev) return;

	ti_prev = ti;
	tim = localtime(&ti);

	_ftprintf(fp, _T("[[ %04d-%02d-%02d %02d:%02d:%02d ]]\n")
		, tim->tm_year + 1900
		, tim->tm_mon + 1
		, tim->tm_mday
		, tim->tm_hour
		, tim->tm_min
		, tim->tm_sec
	);
#else
	SYSTEMTIME sTim;
	GetLocalTime(&sTim);

	_ftprintf(fp, _T("%02d:%02d:%02d.%03d ")
		,sTim.wHour
		,sTim.wDay
		,sTim.wSecond
		,sTim.wMilliseconds
	);
#endif
}
#endif

void MC6809DASM::set_phyaddr(uint32 addr)
{
#ifdef EMABLE_DISASSEMBLER
	phyaddr_tmp = addr;
#endif
}

void MC6809DASM::set_time(uint32 cur)
{
#ifdef EMABLE_DISASSEMBLER
	cur_time_prev = cur_time;
	cur_time = cur;
	cur_time_sum = cur_time - cur_time_prev;

	if (fp == NULL) return;
	_ftprintf(fp, _T("<< %d %05X %04X >>\n"), cur_time_sum, phyaddr, inpc);
#endif
}

void MC6809DASM::ini_pc(uint16 pc, uint8 code)
{
#ifdef EMABLE_DISASSEMBLER
	bool match = false;
	for(int i=0; i<MC6809DASM_PCSTACK_COUNT; i++) {
		if (pc == pcstack[i]) {
			match = true;
			loop_neg_count--;
			break;
		}
	}
	if (!match) {
		loop_neg_count = MC6809DASM_PCSTACK_COUNT;
		loop_started = false;
	}
	pcstack[pcstackpos] = pc;
	pcstackpos = ((pcstackpos+1) % MC6809DASM_PCSTACK_COUNT);

#if 0
	if (pc == 0xeef4) {
		// nmi
		if (triggeron < 100) triggeron += 1000;
	} else if (pc >= 0xf000 && pc < 0xf200) {
		if (triggeron < 1000) triggeron += 1000;
		loop_neg_count = MC6809DASM_PCSTACK_COUNT;
		loop_started = false;
	} else if (pc == 0xfc9e && ina == 0xaf) {
		// keyboard input RETURN key pushed
		if (triggeron < 100) triggeron += 1000;
	} else if (pc == 0xa220) {
		// loop unlock
		if (triggeron < 100) triggeron += 1000;
		loop_neg_count = MC6809DASM_PCSTACK_COUNT;
		loop_started = false;
	} else if (pc == 0xeeee) {
		// firq
		if (triggeron > 100) triggeron = 100;
	} else if (pc == 0xeef1) {
		// irq
		if (triggeron < 100) triggeron += 1000;
	} else if (sc_cat == 0x47) {
		// floppy system call 
		if (triggeron < 100) triggeron += 1000;
	} else if (pc == 0xa532) {
		// floppy read
		if (triggeron < 100) triggeron += 100;
		loop_neg_count = MC6809DASM_PCSTACK_COUNT;
		loop_started = false;
	} else if (pc == 0xa548 || pc == 0xa54c || pc == 0xa553) {
		// floppy write
		if (triggeron < 100) triggeron += 100;
		loop_neg_count = MC6809DASM_PCSTACK_COUNT;
		loop_started = false;
	} else if (phyaddr_tmp >= 0x02000 && phyaddr_tmp < 0x05000) {
		// floppy boot
		if (triggeron < 100) triggeron += 1000;
	} else if (phyaddr_tmp == 0x0723D || phyaddr_tmp == 0x073A3) {
		// floppy
		if (triggeron < 100) triggeron += 1000;
	} else if (pc == 0x8102) {
		// color plette
		if (triggeron < 100) triggeron += 1000;
	}
#endif

	phyaddr = phyaddr_tmp;
	inpc = pc;
	opcode[0] = code;
	opcodelen = 1;
	codelen = 1;
	codepse = 0;
	opmode = NONE;
	memset(&memdata, 0, sizeof(memdata));
	if (triggeron >= 0) triggeron--;
#endif
}

void MC6809DASM::set_mem(uint16 addr, uint8 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	memdata.addr = addr;
	memdata.size = 1;
	memdata.data = data;
	memdata.write = write;
#endif
}

void MC6809DASM::set_mem(uint16 addr, uint16 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	memdata.addr = addr;
	memdata.size = 2;
	memdata.data = data;
	memdata.write = write;
#endif
}

void MC6809DASM::set_phymem(uint32 addr, uint8 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	memdata.phyaddr = addr;

	// trace
	if (addr == 0xeffd1) {
		triggeron += 50;
	}
//	if (addr >= 0xefe24 && addr <= 0xefe2f) {
//		triggeron += 2;
//	} else if (addr == 0xeffd9 && write == true) {
//		triggeron += 10000;
//	}
//	if (addr == 0x0fffe || addr == 0xefffe || addr == 0xffffe) {
//		if (triggeron < 100) triggeron += 100;
//	}
//	if (addr == 0xefe40 || addr == 0xefe41) {
//		if (triggeron < 50) triggeron += 50;
//	}
//	if (addr == 0xeff10 && write) {
//		if (triggeron < 1000) triggeron += 1000;
//	}
#endif
}

void MC6809DASM::set_code2(uint8 code2)
{
#ifdef EMABLE_DISASSEMBLER
	opcode[codelen] = code2;
	opcodelen++;
	codelen++;
#endif
}

void MC6809DASM::set_swi2_param(uint8 cat, uint8 idx)
{
#ifdef EMABLE_DISASSEMBLER
	sc_cat = cat;
	sc_idx = idx;

//	if (sc_cat == 0x41 && (sc_idx >= 0x27 && sc_idx <= 0x28)) {
//		// sin cos
//		if (triggeron < 100) triggeron += 1000;
//	}
#endif
}

void MC6809DASM::set_inh()
{
#ifdef EMABLE_DISASSEMBLER
	opmode = INH;
#endif
}

void MC6809DASM::set_swi()
{
#ifdef EMABLE_DISASSEMBLER
	opmode = INH;
#endif
}

void MC6809DASM::set_swi2()
{
#ifdef EMABLE_DISASSEMBLER
	opmode = INH;
	swi2fire = 1;
//	triggeron = 100;
#endif
}

void MC6809DASM::set_swi3()
{
#ifdef EMABLE_DISASSEMBLER
	opmode = INH;
#endif
}

#ifdef EMABLE_DISASSEMBLER
void MC6809DASM::set_dir_addr(uint8 dp, uint8 addrl)
{
	opmode = DIR;
	opcode[codelen] = addrl;
	codelen++;
	indp = dp;
}
#endif

void MC6809DASM::set_dir(uint8 dp, uint8 addrl, uint8 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	set_dir_addr(dp, addrl);
	set_mem((dp << 8) | addrl, data, write);
#endif
}

void MC6809DASM::set_dir(uint8 dp, uint8 addrl, uint16 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	set_dir_addr(dp, addrl);
	set_mem((dp << 8) | addrl, data, write);
#endif
}

void MC6809DASM::set_imm(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = IMM1;
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_imm(uint16 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = IMM2;
	opcode[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	opcode[codelen] = (data & 0xff);
	codelen++;
#endif
}

#ifdef EMABLE_DISASSEMBLER
void MC6809DASM::set_ext_addr(uint16 addr)
{
	opmode = EXT;
	opcode[codelen] = ((addr & 0xff00) >> 8);
	codelen++;
	opcode[codelen] = (addr & 0xff);
	codelen++;
}
#endif

void MC6809DASM::set_ext(uint16 addr, uint8 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	set_ext_addr(addr);
	set_mem(addr, data, write);
#endif
}

void MC6809DASM::set_ext(uint16 addr, uint16 data, bool write)
{
#ifdef EMABLE_DISASSEMBLER
	set_ext_addr(addr);
	set_mem(addr, data, write);
#endif
}

void MC6809DASM::set_idx1(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = IDX;
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_idx2(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_idx2(uint16 data)
{
#ifdef EMABLE_DISASSEMBLER
	opcode[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	opcode[codelen] = (data & 0xff);
	codelen++;
#endif
}

void MC6809DASM::set_rel(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = REL1;
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_rel(uint16 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = REL2;
	opcode[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	opcode[codelen] = (data & 0xff);
	codelen++;
#endif
}

void MC6809DASM::set_jmp(uint16 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = JMP;
	opcode[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	opcode[codelen] = (data & 0xff);
	codelen++;
#endif
}

void MC6809DASM::set_tfr(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = TFR;
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_psh(uint8 data)
{
#ifdef EMABLE_DISASSEMBLER
	opmode = PSH;
	opcode[codelen] = data;
	codelen++;
#endif
}

void MC6809DASM::set_err()
{
#ifdef EMABLE_DISASSEMBLER
	opmode = ERR;
#endif
}

void MC6809DASM::set_reg(uint8 cc, uint8 dp, uint8 a, uint8 b, uint16 x, uint16 y, uint16 s, uint16 u)
{
#ifdef EMABLE_DISASSEMBLER
	incc = cc;
	indp = dp;
	ina = a;
	inb = b;
	inx = x;
	iny = y;
	ins = s;
	inu = u;
#endif
}

void MC6809DASM::print(int accum, int cycles)
{
#ifdef EMABLE_DISASSEMBLER
	if (triggeron > 0 && loop_neg_count > 0) {
		_stprintf(line, _T("%07d %05X %04X"), accum, phyaddr, inpc);
		for(int i=0; i < 5; i++) {
			if (i < codelen) {
				_stprintf(&line[_tcslen(line)], _T(" %02X"), opcode[i]);
			} else {
				_tcscat(line, _T("   "));
			}
		}

		codepse = 0;
		set_cmd_str();
		switch(opmode) {
		case INH:
			break;
		case DIR:
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
				, indp, opcode[codepse]
			);
			break;
		case EXT:
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
				, opcode[codepse], opcode[codepse+1]
			);
			break;
		case IMM1:
			_stprintf(&cmd[_tcslen(cmd)], _T(" #$%02X")
				, opcode[codepse]
			);
			break;
		case IMM2:
			_stprintf(&cmd[_tcslen(cmd)], _T(" #$%02X%02X")
				, opcode[codepse], opcode[codepse+1]
			);
			break;
		case REL1:
			{
				uint16 data = opcode[codepse];
				data = inpc + 1 + opcodelen + ((data & 0x80) ? data - 256 : data);
				_stprintf(&cmd[_tcslen(cmd)], _T(" J%04X")
					, data 
				);
			}
			break;
		case REL2:
			{
				uint16 data = (opcode[codepse] << 8) + opcode[codepse + 1];
				data = inpc + 2 + opcodelen + ((data & 0x8000) ? data - 65536 : data);
				_stprintf(&cmd[_tcslen(cmd)], _T(" J%04X")
					, data
				);
			}
			break;
		case JMP:
			_stprintf(&cmd[_tcslen(cmd)], _T(" J%02X%02X")
				, opcode[codepse], opcode[codepse+1]
			);
			break;
		case IDX:
			set_idx_str();
			break;
		case TFR:
			set_tfr_str();
			break;
		case PSH:
			set_psh_str();
			break;
		default:
			break;
		}
		_stprintf(&line[_tcslen(line)], _T(" %-18s"), cmd);

		// register
		_stprintf(&line[_tcslen(line)], _T("  [%2d](%c%c%c%c%c%c%c%c) DP:%02X A:%02X B:%02X X:%04X Y:%04X S:%04X U:%04X")
			,cycles
			,(incc & 0x80) ? _T('E') : _T(' ')
			,(incc & 0x40) ? _T('F') : _T(' ')
			,(incc & 0x20) ? _T('H') : _T(' ')
			,(incc & 0x10) ? _T('I') : _T(' ')
			,(incc & 0x08) ? _T('N') : _T(' ')
			,(incc & 0x04) ? _T('Z') : _T(' ')
			,(incc & 0x02) ? _T('V') : _T(' ')
			,(incc & 0x01) ? _T('C') : _T(' ')
			,indp
			,ina
			,inb
			,inx
			,iny
			,ins
			,inu
		);

		// memory data
		if (memdata.size == 1) {
			_stprintf(&line[_tcslen(line)], _T("  (%c %05X %04X:%02X)"), memdata.write ? _T('W') : _T('R'), memdata.phyaddr, memdata.addr, memdata.data & 0xff);
		} else if (memdata.size == 2) {
			_stprintf(&line[_tcslen(line)], _T("  (%c %05X %04X:%04X)"), memdata.write ? _T('W') : _T('R'), memdata.phyaddr, memdata.addr, memdata.data);
		}

		if (fp != NULL) {
//			print_localtime();
			_ftprintf(fp, _T("%s\n"), line);
		}
	}
	if (swi2fire) {
		if (triggeron > 0 && loop_neg_count > 0) {
			_stprintf(line, _T("        Category:%02X Index:%02X"), sc_cat, sc_idx);
			if (fp != NULL) {
				_ftprintf(fp, _T("%s\n"), line);
			}
		}
		sc_cat = 0;
		sc_idx = 0;
		swi2fire = 0;
	}
	if (triggeron == 0 && loop_neg_count > 0) {
		if (fp != NULL) {
			_ftprintf(fp, _T("        (Logging off)\n"));
		}
	}
	if (!loop_started && loop_neg_count < 0) {
		if (triggeron > 0) {
			if (fp != NULL) {
				_ftprintf(fp, _T("        (Loop...)\n"));
			}
		}
		loop_started = true;
	}
#endif
}

void MC6809DASM::print_signal(int id, uint32 data, int accum)
{
#ifdef EMABLE_DISASSEMBLER
	if (fp == NULL) return;
	uint16 now_intr = (data & 0xffff);
	switch(id) {
	case SIG_CPU_IRQ:
		if (triggeron > 0 && now_intr != now_irq) {
//			print_localtime();
			_ftprintf(fp, _T("%07d >>>> IRQ (%d) Interrupt %x\n"), accum, id, now_intr);
		}
		now_irq = now_intr;
		break;
	case SIG_CPU_FIRQ:
		if (triggeron > 0 && now_intr != now_firq) {
//			print_localtime();
			_ftprintf(fp, _T("%07d >>>> FIRQ (%d) Interrupt %x\n"), accum, id, now_intr);
		}
		now_firq = now_intr;
		break;
	case SIG_CPU_NMI:
		if (now_intr != now_nmi) {
//			print_localtime();
			_ftprintf(fp, _T("%07d >>>> NMI (%d) Interrupt %x\n"), accum, id, now_intr);
		}
		now_nmi = now_intr;
		break;
	case SIG_CPU_RESET:
		if (now_intr != now_reset) {
//			print_localtime();
			_ftprintf(fp, _T("%07d >>>> RES (%d) Interrupt %x\n"), accum, id, now_intr);
		}
		now_reset = now_intr;
		break;
	case SIG_CPU_HALT:
		if (now_intr != now_halt) {
//			print_localtime();
//			_ftprintf(fp, _T("%07d >>>> HALT (%d) Interrupt %x\n"), accum, id, now_intr);
		}
		now_halt = now_intr;
		break;
	}
#endif
}

void MC6809DASM::print_intr(int id, int accum, int cycles)
{
#ifdef EMABLE_DISASSEMBLER
	if (fp == NULL) return;
	switch(id) {
	case SIG_CPU_IRQ:
//		print_localtime();
		if (triggeron > 0) _ftprintf(fp, _T("%07d >>>> IRQ (%d) fired [%2d]\n"), accum, id, cycles);
		break;
	case SIG_CPU_FIRQ:
//		print_localtime();
		if (triggeron > 0) _ftprintf(fp, _T("%07d >>>> FIRQ (%d) fired [%2d]\n"), accum, id, cycles);
		break;
	case SIG_CPU_NMI:
//		print_localtime();
		_ftprintf(fp, _T("%07d >>>> NMI (%d) fired [%2d]\n"), accum, id, cycles);
//		if (triggeron < 100) triggeron += 100;
		break;
	}
#endif
}

void MC6809DASM::trace_on(int val)
{
#ifdef EMABLE_DISASSEMBLER
	triggeron += val;
#endif
}

#ifdef EMABLE_DISASSEMBLER

void MC6809DASM::set_cmd_str()
{
	uint8 ch = opcode[0];
	uint8 cl = opcode[1];

	memset(cmd, 0, sizeof(cmd));

	if (ch == 0x10) {
		for(int i=0; opcodes1[i].s != NULL; i++) {
			if (cl == opcodes1[i].c) {
				_stprintf(cmd, _T("%-5s"), opcodes1[i].s);
				codepse = 2;
				break;
			}
		}
		if (codepse != 2) {
			_stprintf(cmd, _T("%-5s"), _T("??"));
		}
	} else if (ch == 0x11) {
		for(int i=0; opcodes2[i].s != NULL; i++) {
			if (cl == opcodes2[i].c) {
				_stprintf(cmd, _T("%-5s"), opcodes2[i].s);
				codepse = 2;
				break;
			}
		}
		if (codepse != 2) {
			_stprintf(cmd, _T("%-5s"), _T("??"));
		}
	} else {
		_stprintf(cmd, _T("%-5s"), opcodes0[ch].s);
		codepse = 1;
	}
}

void MC6809DASM::set_idx_str()
{
	_TCHAR reg;

	uint8 data = opcode[codepse];
	uint8 pdata1 = opcode[codepse+1];
	uint16 pdata2 = ((opcode[codepse+1] << 8) | opcode[codepse+2]);

	switch(data & 0x60) {
	case 0x00: reg = _T('X'); break;
	case 0x20: reg = _T('Y'); break;
	case 0x40: reg = _T('U'); break;
	case 0x60: reg = _T('S'); break;
	}
	if ((data & 0x80) == 0) {
		// 0x00 - 0x7f
		_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
		, ((int)data & 0xf) - ((data & 0x10) ? 16 : 0)
		, reg);
	} else if ((data & 0x8f) == 0x80) {
		_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c+")
		, reg);
	} else if ((data & 0x8f) == 0x81) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c++")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,%c++]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x82) {
		_stprintf(&cmd[_tcslen(cmd)], _T(" ,-%c")
		, reg);
	} else if ((data & 0x8f) == 0x83) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,--%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,--%c]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x84) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,%c]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x85) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" B,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [B,%c]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x86) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" A,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [A,%c]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x88) {
		// 8bit offset
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
			, (int)pdata1 - ((pdata1 & 0x80) ? 256 : 0)
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [%d,%c]")
			, (int)pdata1 - ((pdata1 & 0x80) ? 256 : 0)
			, reg);
		}
	} else if ((data & 0x8f) == 0x89) {
		// 16bit offset
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
			, (int)pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [%d,%c]")
			, (int)pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)
			, reg);
		}
	} else if ((data & 0x8f) == 0x8b) {
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" D,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [D,%c]")
			, reg);
		}
	} else if ((data & 0x8f) == 0x8c) {
		// 8bit offset PCR
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X,PCR")
			, (codepse + 2 + inpc + pdata1 - ((pdata1 & 0x80) ? 256 : 0)) & 0xffff
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X,PCR]")
			, (codepse + 2 + inpc + pdata1 - ((pdata1 & 0x80) ? 256 : 0)) & 0xffff
			, reg);
		}
	} else if ((data & 0x8f) == 0x8d) {
		// 16bit offset PCR
		if ((data & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X,PCR")
			, (codepse + 3  + inpc + pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)) & 0xffff
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X,PCR]")
			, (codepse + 3  + inpc + pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)) & 0xffff
			, reg);
		}
	} else if ((data & 0x9f) == 0x9f) {
		// extended indirect
		_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X]")
		, pdata2 & 0xffff
		, reg);
	}
}

static const _TCHAR *tfr_regs[] = {
	_T("D"),
	_T("X"),
	_T("Y"),
	_T("U"),
	_T("S"),
	_T("PC"),
	_T("?"),
	_T("?"),

	_T("A"),
	_T("B"),
	_T("CC"),
	_T("DP"),
	_T("?"),
	_T("?"),
	_T("?"),
	_T("?"),
	NULL
};

void MC6809DASM::set_tfr_str()
{
	uint8 dh = ((opcode[codepse] & 0xf0) >> 4);
	uint8 dl = (opcode[codepse] & 0xf);

	_stprintf(&cmd[_tcslen(cmd)], _T(" %s,%s")
		, tfr_regs[dh], tfr_regs[dl]
	);
}

static const _TCHAR *psh_regs[] = {
	_T("CC"),
	_T("A"),
	_T("B"),
	_T("DP"),
	_T("X"),
	_T("Y"),
	_T("S/U"),
	_T("PC"),
	NULL
};

void MC6809DASM::set_psh_str()
{
	uint8 data = opcode[codepse];
	bool is_s = ((opcode[0] & 0xf) < 0x06); 
	bool one = true;
	for(int i=0; i < 8; i++) {
		if ((data & (1 << i)) != 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T("%c%s")
				, one ? _T(' ') : _T(',')
				, (i == 6 ? (is_s ? _T("U") : _T("S")) : psh_regs[i])
			);
			one = false;
		}
	}
}

#endif /* ENABLE_DISASSEMBLER */

//for dis-assemble
MC6809DASM dasm;
