/** @file mc6809dasm.h

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.08.29 -

	@brief [ mc6809 disassembler ]
*/

#ifndef _MC6809DASM_H_
#define _MC6809DASM_H_

#include "../common.h"
#include <stdio.h>

#define MC6809DASM_PCSTACK_COUNT 20

class MC6809DASM
{
private:
#ifdef EMABLE_DISASSEMBLER
	enum en_opmode {
		NONE = 0,
		INH,
		DIR,
		EXT,
		IMM1,
		IMM2,
		IDX,
		REL1,
		REL2,
		JMP,
		TFR,
		PSH,
		ERR,
	};

	_TCHAR line[_MAX_PATH];
	uint32 phyaddr;
	uint32 phyaddr_tmp;
	uint16 inpc;
	uint8  opcode[8];
	int    opcodelen;
	int    codelen;
	int    codepse;
	en_opmode opmode;
	struct st_memdata {
		uint32 phyaddr;
		uint16 addr;
		uint8  size;
		uint16 data;
		bool   write;
	} memdata;

	uint8  indp;
	uint8  incc;
	uint8  ina;
	uint8  inb;
	uint16 inx;
	uint16 iny;
	uint16 ins;
	uint16 inu;

	_TCHAR cmd[128];

	// start logging trigger
	int    triggeron;
	// for system call
	uint8  swi2fire;
	uint8  sc_cat, sc_idx;
	// for loop detection
	uint16 pcstack[MC6809DASM_PCSTACK_COUNT];
	int    pcstackpos;
	int    loop_neg_count;
	bool   loop_started;

	uint16 now_halt;
	uint16 now_reset;
	uint16 now_nmi;
	uint16 now_irq;
	uint16 now_firq;

	FILE *fp;
	time_t ti,ti_prev;
	struct tm *tim;

	uint32 cur_time, cur_time_prev;
	int cur_time_sum;

	void set_cmd_str();
	void set_idx_str();
	void set_tfr_str();
	void set_psh_str();

	void set_ext_addr(uint16 addr);
	void set_dir_addr(uint8 dp, uint8 addrl);

	void print_localtime();

#endif /* ENABLE_DISASSEMBLER */

public:
	MC6809DASM();
	~MC6809DASM();

	bool open();
	void close();

	void set_phyaddr(uint32 addr);
	void set_time(uint32 cur);
	void ini_pc(uint16 pc, uint8 code);
	void set_code2(uint8 code2);
	void set_swi2_param(uint8 cat, uint8 idx);

	void set_mem(uint16 addr, uint8 data, bool write = false);
	void set_mem(uint16 addr, uint16 data, bool write = false);
	void set_phymem(uint32 addr, uint8 data, bool write = false);

	void set_inh();
	void set_swi();
	void set_swi2();
	void set_swi3();

	void set_dir(uint8 dp, uint8 addrl, uint8 data, bool write = false);
	void set_dir(uint8 dp, uint8 addrl, uint16 data, bool write = false);

	void set_imm(uint8 data);
	void set_imm(uint16 data);

	void set_ext(uint16 addr, uint8 data, bool write = false);
	void set_ext(uint16 addr, uint16 data, bool write = false);

	void set_idx1(uint8 data);
	void set_idx2(uint8 data);
	void set_idx2(uint16 data);

	void set_rel(uint8 data);
	void set_rel(uint16 data);
	void set_jmp(uint16 data);

	void set_tfr(uint8 data);

	void set_psh(uint8 data);

	void set_err();

	void set_reg(uint8 cc, uint8 dp, uint8 a, uint8 b, uint16 x, uint16 y, uint16 s, uint16 u);

	void print(int accum, int cycles);

	void print_signal(int id, uint32 data, int accum);

	void print_intr(int id, int accum, int cycles);

	void trace_on(int val);
};

extern MC6809DASM dasm;

#endif /* _MC6809DASM_H_ */
